/* 

  speed 0.1 - (C) by Bernhard Oemer <oemer@tph.tuwien.ac.at> 1999

  This program is free software according to the GPL.

  BUILD: gcc speed.c -o speed

  USAGE: speed [options] [input file/device]
  measures the throughput form a file, device or a pipe.
  -h, --help           display this message.
  -V, --version        display version and copyright message and exit.
  -o, --log=<logfile>  copy output to <logfile>.
  -q, --quiet          don't write output to stderr.
  -v, --verbose        verbose output headers.
  -e, --echo           copy input to stdout (for use within a pipe).
  -m, --min            report minimum rate in sample.
  -x, --max            report maximum rate in sample.
  -b, --bufsize=<n>    set size of internal buffer to n chars.
  -s, --samples=<n>    subdevide sample period into n samples.
  -p, --period=<n>     set sample period to n chars.
  -n, --number=<n>     max. number of measurements periods.
  -c, --scale=<n>      display bars of '#'s, one '#' = n chars.
  all numeric values accept the postfix multipliers 'K' and 'M'.

  EXAMPLES:
  
  Test your filesystem throughput for CD-writing:
  
    $ mkisofs -R /pub 2> /dev/null | speed 

  Benchmark your harddisk and scan for defect regions:
  
    $ speed -v -c 256k -p 20m -s 10 -m /dev/hda2
  
  Monitor entropy generating hardware events:   
  
    $ speed -c 10 -p 128 /dev/random

*/ 


#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <limits.h>
#include <signal.h>

char short_options[] = "hVo:qvemxb:s:p:n:c:";

struct option long_options[] = {
  { "help",0,0,'h' },
  { "version",0,0,'V' },
  { "log",1,0,'o' },
  { "quiet",0,0,'q' },
  { "verbose",0,0,'v' },
  { "echo",0,0,'e' },
  { "min",0,0,'m' },
  { "max",0,0,'x' },
  { "bufsize",1,0,'b' },
  { "samples",1,0,'s' },
  { "period",1,0,'p' },
  { "number",1,0,'n' },
  { "scale",1,0,'c' },
  { 0,0,0,0 }
};

char usage[]=
  "USAGE: %s [options] [input file/device]\n"
  "measures the throughput form a file, device or a pipe.\n"
  "-h, --help           display this message.\n"
  "-V, --version        display version and copyright message and exit.\n"
  "-o, --log=<logfile>  copy output to <logfile>.\n"
  "-q, --quiet          don't write output to stderr.\n"
  "-v, --verbose        print output header and stats.\n"
  "-e, --echo           copy input to stdout (for use within a pipe).\n"
  "-m, --min            report minimum rate in sample.\n" 
  "-x, --max            report maximum rate in sample.\n" 
  "-b, --bufsize=<n>    set size of internal buffer to n chars.\n"
  "-s, --samples=<n>    subdevide sample period into n samples.\n"
  "-p, --period=<n>     set sample period to n chars.\n"
  "-n, --number=<n>     max. number of measurements periods.\n"
  "-c, --scale=<n>      display bars of '#'s, one '#' = n chars.\n"
  "all numeric values accept the postfix multipliers 'K' and 'M'.\n"
;

char version[]=
  "%s version 0.1 - (c) by Bernhard Oemer <oemer@tph.tuwien.ac.at> 1999\n\n"
  "This program is free software; you can redistribute it and/or modify\n"
  "it under the terms of the GNU General Public License as published by\n"
  "the Free Software Foundation; either version 2 of the License, or\n"
  "(at your option) any later version.\n\n"
  "This program is distributed in the hope that it will be useful,\n"
  "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
  "GNU General Public License for more details.\n"
;  

#define STAT_AVG 0
#define STAT_MIN 1
#define STAT_MAX 2

char *stat_str[]={ " [average]"," [minimum]"," [maximum]" };

#define UNDEF (-1)
#define KB *1024
#define MB *1024*1024
#define BUFSIZE 64 KB

FILE *log=NULL;
FILE *f=stdin;
int quiet=0;
int verbose=0;
int echo=0;
int stat=STAT_AVG;
int bufsize=UNDEF;
int samples=1;
int period=10 MB;
int number=INT_MAX;
int scale=0;
clock_t clock0;
long long chars_read;
int term_req=0;

int argtol(char *s) {
  char *p;
  int n;
  
  n=strtol(s,&p,10);
  if(*p == 'k' || *p == 'K') { n*=1024; p++; }
  if(*p == 'm' || *p == 'M') { n*=1024*1024; p++; }
  if(*p || n<=0) {
    fprintf(stderr,"invalid numeric parameter.\n");
    exit(1);
  };
  return n;
}

void printlog(char *s) {
  if(!quiet) fputs(s,stderr);
  if(log) fputs(s,log);
}

double secs() {
  return (double)(times(0)-clock0)/(double)CLOCKS_PER_SEC;
}

void sighandler(int sig) {
  term_req=1;
} 

main(int argc,char **argv) {
  int n,i,j,c;
  char *buf;
  char line[256];
  double t0,t1,t,rate;
  double blocks_per_buf,bufs_per_sample,kb_per_period,mb_per_period;
  size_t s;
    
  while(1) {
    c=getopt_long(argc,argv,short_options,long_options,&i);
    if(c==-1) break;
    switch(c) {
      case 'h':
        fprintf(stderr,usage,argv[0]);
        exit(0);
      case 'V':
        fprintf(stderr,version,argv[0]);
        exit(0);
      case 'o':
        log=fopen(optarg,"w");
        if(!log) {
          fprintf(stderr,"can't open %s for output.\n",optarg);
          exit(1);
        };
        break;
      case 'q': quiet=1; break;
      case 'v': verbose=1; break;
      case 'e': echo=1; break;
      case 'm': stat=STAT_MIN; break;
      case 'x': stat=STAT_MAX; break;
      case 'b': bufsize=argtol(optarg); break;
      case 's': samples=argtol(optarg); break;
      case 'p': period=argtol(optarg); break;
      case 'n': number=argtol(optarg); break;
      case 'c': scale=argtol(optarg); break;
      case ':': 
        fprintf(stderr,"missing parameter to option -%c.\n",c);
        exit(1);
      default:
        fprintf(stderr,"illegal option, use %s -h for help.\n",argv[0]);
        exit(1);
    }
  }
  if(stat==STAT_AVG) samples=1;
  if(bufsize==UNDEF) {
    bufsize=period<samples*BUFSIZE ? bufsize=period/samples : BUFSIZE;
    while(period%bufsize) bufsize/=2;
  } 
  if(period%(bufsize*samples)) {
    fprintf(stderr,"invalid sample period.\n",c);
    exit(1);
  }      
  if(argc>optind+1) {
    fprintf(stderr,"only one input file/device can be specified.\n",c);
    exit(1);
  }      
  buf=malloc(bufsize);
  if(!buf) {      
    fprintf(stderr,"can't allocate buffer.\n",c);
    exit(1);
  }
  if(argc==optind+1) {
    f=fopen(argv[optind],"r");
    if(!f) {      
      fprintf(stderr,"can't open input file/device %s.\n",argv[optind]);
      exit(1);
    };
  }

  if(verbose) {
    sprintf(line,"file/device: %s, bufsize=%d, period=%d, samples=%d%s\n",
             (f == stdin ? "pipe" : argv[optind]),bufsize,period,samples,
             (samples > 1 ? stat_str[stat] : "")
           );
    printlog(line);
  }
  bufs_per_sample=period/(bufsize*samples);
  kb_per_period=(double)period/1024.0;
  mb_per_period=(double)period/(1024.0*1024.0);
  clock0=times(0); chars_read=0;
  signal(SIGINT,&sighandler);
  t0=secs();
  for(n=1;n<=number;n++) {
    
    if(stat=STAT_MIN) t=INT_MAX;
    if(stat=STAT_MAX) t=0;
    for(i=0;i<samples;i++) {
      for(j=0;j<bufs_per_sample;j++) {
        s=fread(buf,1,bufsize,f);
        chars_read+=s;
        if(echo) fwrite(buf,1,s,stdout);
        if(ferror(f)) {
          fprintf(stderr,"I/O-Error reading input file/device.\n");
          goto end;
        }
        if(feof(f) || term_req) goto end; 
      }
      t1=secs();
      switch(stat) {
        case STAT_AVG: t=t1-t0; break;
        case STAT_MAX: if(t<t1-t0) t=t1-t0; break;
        case STAT_MIN: if(t>t1-t0) t=t1-t0; break;
        default: exit(1);
      };
      t0=t1;
    }
    rate=t ? period/(t*samples) : 2*CLOCKS_PER_SEC*period;
    if(period>=1 MB) 
      sprintf(line,"%8.1f MB",n*mb_per_period); 
    else if(period>=1 KB)
      sprintf(line,"%8.1f KB",n*kb_per_period);
    else
      sprintf(line,"%7d bytes",n*period);
    printlog(line);
    if(period>=100 KB) 
      sprintf(line,", %6.0f KB/s",rate/(1 KB)); 
    else 
      sprintf(line,", %6.0f bytes/s",rate);
    printlog(line);
      
    i=0;
    if(scale) {
      sprintf(line,": ");
      printlog(line);
      i=rate/scale;
      if(i>100) i=100;
      for(j=0;j<i;j++) line[j]='#';
    };
    line[i]='\n';line[i+1]=0;
    printlog(line);
  }
  
end:

  if(f != stdin) fclose(f);
  t=secs();
  if(verbose) {
    sprintf(line,"%Ld bytes read in %.2f seconds, avg. throughput %.1f KB/s\n",
             chars_read,t,chars_read/(t KB)
           );
    printlog(line); 
  }
  if(log) fclose(log);
  exit(0);
}      

